@@ -317,7 +317,7 @@ main( int argc, char *argv[] )
 	      "malshow            -- Display results from malaga.\n"
 	      "malshow -v[ersion] -- Print the version number.\n"
 	      "malshow -h[elp]    -- Print this help.\n\n"
-	      "This program is usually started by malaga.\n" );
+	      "This program is usually started by malaga or mallex.\n" );
       exit(0);
     }
   }
