@@ -1,4 +1,4 @@
-/* Copyright (C) 1995 Bjoern Beutel. */
+/* Copyright (C) 2002 Bjoern Beutel. */
 
 /* Description. =============================================================*/
 
@@ -6,12 +6,10 @@
 
 /* Includes. ================================================================*/
 
-#define _XOPEN_SOURCE
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
 #include <setjmp.h>
-#include <ctype.h>
 #include <string.h>
 #include <gtk/gtk.h>
 #include "basic.h"
@@ -24,23 +22,28 @@
 
 enum {FIRST_STATE, NEXT_STATE, PREV_STATE, LAST_STATE};
 
+enum {PATH_BEGIN, PATH_END, FROM_ROOT}; /* Items of the tree's pop-up menu. */
+
 /* Types. ===================================================================*/
 
 typedef enum {INTER_NODE, BREAK_NODE, FINAL_NODE, UNFINAL_NODE,
 	      PRUNED_NODE} tree_node_type_t;
 
+typedef enum {ELEMENT_NODE, ELEMENT_LINK, ELEMENT_RULE} element_t;
+/* Elements of a node that may be hit by the mouse pointer. */
+
 typedef struct tree_node 
 { 
   struct tree_node *parent; /* Parent of this tree node. */
   struct tree_node *sibling; /* Next sibling of this tree node. */
   struct tree_node *child; /* First successor of this tree node. */
-  int_t index; /* Index of this tree node. */
+  int_t state_index; /* State index of this tree node. */
   tree_node_type_t type; /* Type of this node. */
   string_t rule_name; /* Name of rule that created this node. */
   string_t link_surf;
-  string_t link_fs;
+  string_t link_feat;
   string_t result_surf;
-  string_t result_fs;
+  string_t result_feat;
   string_t rule_set;
   bool_t result_path;
 
@@ -49,43 +52,53 @@ typedef struct tree_node
   int_t x, y; /* Node position. */
   pos_string_t *pos_rule_name; /* Positioned rule name. */
   pos_string_t *pos_link_surf; /* Positioned link surface. */
+  pos_string_t *pos_state_index; /* Positioned state index. */
 } tree_node_t;
 
 typedef struct
 {
   list_node_t *next; /* Next path node in this list. */
-  pos_string_t *index;
+  pos_string_t *state_index;
   pos_string_t *link_surf;
-  pos_value_t *link_fs;
+  pos_value_t *link_feat;
   pos_string_t *rule_name;
   pos_string_t *result_surf;
-  pos_value_t *result_fs;
+  pos_value_t *result_feat;
   pos_string_t *rule_set;
 } path_node_t;
 
 /* Global variables. ========================================================*/
 
 rectangle_t tree_geometry, path_geometry;
-string_t tree_font_name, path_font_name;
+string_t tree_font_family, path_font_family;
 int_t tree_font_size, path_font_size;
+tree_mode_t tree_mode;
+bool_t show_state_indexes;
+bool_t inline_path;
 
 /* Variables. ===============================================================*/
 
 /* Information in Tree window. */
 static pos_string_t *result_surf;
 static tree_node_t *tree_nodes;
-static tree_node_t **nodes; /* A dynamic array of node pointers. */
-static int_t node_count; /* Size of the array NODES. */
-static int_t width, height; /* Width and height of tree window. */
+static tree_node_t **states; /* Dynamic array of nodes representing states. */
+static int_t state_count; /* Size of the array STATES. */
+static int_t tree_width, tree_height; /* Width and height of tree canvas. */
 static canvas_t *tree_canvas;
-static enum {FULL_TREE, NO_DEAD_ENDS, RESULT_PATHS} tree_mode;
+static tree_node_t *popup_menu_node;
+
+static bool_t tree_show_state_indexes; 
+/* Controls whether state indexes are shown in tree. */
 
 /* Information in Path window. */
 static list_t path_nodes;
 static canvas_t *path_canvas;
 static pos_string_t *plus;
+static pos_string_t *comma;
 static tree_node_t *path_begin, *path_end;
-static bool_t in_line; /* If TRUE, feature structures are shown in line. */
+
+static bool_t path_show_state_indexes; 
+/* Controls whether state indexes are shown in path. */
 
 /* Forward declarations. ====================================================*/
 
@@ -133,7 +146,7 @@ configure_path( canvas_t *canvas, int_t *width_p, int_t *height_p )
 /* Do the layout of the path CANVAS. 
  * Return the canvas' total size in *WIDTH_P and *HEIGHT_P. */
 {
-  int_t width, height;
+  int_t path_width, path_height, x, x_dist;
   path_node_t *node;
   bool_t is_first;
   int_t space_width = get_space_width( canvas );
@@ -142,101 +155,125 @@ configure_path( canvas_t *canvas, int_t *width_p, int_t *height_p )
   int_t border_width = get_border_width( canvas );
   
   config_pos_string( plus, canvas );
+  config_pos_string( comma, canvas );
+  x_dist = space_width + comma->width;
 
-  width = height = border_width;
+  path_width = path_height = border_width;
   is_first = TRUE;
 
   FOREACH( node, path_nodes )
   {
-    if (node->link_fs != NULL)
+    if (node->link_feat != NULL)
     {
       if (is_first) 
 	is_first = FALSE; 
       else 
-	height += font_height;
+	path_height += font_height;
       
       config_pos_string( node->link_surf, canvas );
       node->link_surf->x = border_width + plus->width + space_width;
-      width = MAX( width, node->link_surf->x + node->link_surf->width );
+      path_width = MAX( path_width, 
+			node->link_surf->x + node->link_surf->width );
       
-      config_pos_value( node->link_fs, canvas );
-      node->link_fs->x = node->link_surf->x;
-      if (in_line)
+      config_pos_value( node->link_feat, canvas );
+      node->link_feat->x = node->link_surf->x;
+      if (inline_path)
       {
-	node->link_surf->y = height + node->link_fs->ascent - font_ascent;
-	node->link_fs->x += node->link_surf->width + 2 * space_width;
+	node->link_surf->y = (path_height 
+			      + node->link_feat->ascent - font_ascent);
+	node->link_feat->x += node->link_surf->width + x_dist;
       }
       else
       {
-	node->link_surf->y = height;
-	height += font_height;
+	node->link_surf->y = path_height;
+	path_height += font_height;
       }
-      node->link_fs->y = height;
-      width = MAX( width, node->link_fs->x + node->link_fs->width );
-      height += node->link_fs->height;
+      node->link_feat->y = path_height;
+      path_width = MAX( path_width, 
+			node->link_feat->x + node->link_feat->width );
+      path_height += node->link_feat->height;
     }
-    if (node->result_fs != NULL)
+    if (node->result_feat != NULL)
     {
-      if (is_first) 
+      if (is_first)
 	is_first = FALSE; 
-      else 
-	height += font_height;
+      else
+	path_height += font_height;
 
-      config_pos_string( node->rule_name, canvas );
-      node->rule_name->x = border_width;
-      node->rule_name->y = height;
-      config_pos_string( node->index, canvas );
-      node->index->x = (node->rule_name->x + node->rule_name->width 
-			+ 2 * space_width);
-      node->index->y = height;
+      x = border_width;
+
+      /* Configure rule name. */
+      if (node != (path_node_t *) path_nodes.first)
+      {
+	config_pos_string( node->rule_name, canvas );
+	node->rule_name->x = x;
+	x += node->rule_name->width + 2 * space_width;
+	if (! inline_path)
+	  node->rule_name->y = path_height;
+      }
+
+      /* Configure state index. */
+      if (path_show_state_indexes)
+      {
+	config_pos_string( node->state_index, canvas );
+	node->state_index->x = x;
+	if (inline_path)
+	  x += node->state_index->width + x_dist;
+	else
+	{
+	  node->state_index->y = path_height;
+	  path_width = MAX( path_width, x + node->state_index->width );
+	  path_height += font_height;
+	}
+      }
+
+      /* Configure result surface. */
       config_pos_string( node->result_surf, canvas );
-      node->result_surf->x = (node->index->x + node->index->width 
-			      + 2 * space_width);
-      node->result_surf->y = height;
-      width = MAX( width, node->result_surf->x + node->result_surf->width );
-      if (! in_line) 
-	height += font_height;
+      node->result_surf->x = x;
+      if (inline_path) 
+	x += node->result_surf->width + x_dist;
+      else
+      {
+	node->result_surf->y = path_height;
+	path_width = MAX( path_width, x + node->result_surf->width );
+	path_height += font_height;
+      }
 
       /* Configure result feature structure. */
-      config_pos_value( node->result_fs, canvas );
-      node->result_fs->y = height;
-      if (in_line)
-      {
-	node->result_fs->x = (node->result_surf->x + node->result_surf->width
-			      + 2 * space_width);
-      }
+      config_pos_value( node->result_feat, canvas );
+      node->result_feat->x = x;
+      node->result_feat->y = path_height;
+      if (inline_path)
+	x += node->result_feat->width + x_dist;
       else
       {
-	node->result_fs->x = node->result_surf->x;
-	height += node->result_fs->height;
+	path_width = MAX( path_width, x + node->result_feat->width );
+	path_height += node->result_feat->height;
       }
-      width = MAX( width, node->result_fs->x + node->result_fs->width );
 
       /* Configure rule set. */
       config_pos_string( node->rule_set, canvas );
-      node->rule_set->y = height;
-      if (in_line)
+      node->rule_set->x = x;
+      if (inline_path)
       {
-	node->rule_set->x = (node->result_fs->x +node->result_fs->width 
-			     + 2 * space_width);
 	node->rule_name->y 
-	  = node->index->y 
+	  = node->state_index->y 
 	  = node->result_surf->y 
 	  = node->rule_set->y
-	  = height + node->result_fs->ascent - font_ascent;
-	height += node->result_fs->height;
+	  = path_height + node->result_feat->ascent - font_ascent;
+	path_height += node->result_feat->height;
       }
       else
       {
-	node->rule_set->x = node->result_surf->x;
-	height += font_height;
+	node->rule_set->y = path_height;
+	path_height += font_height;
       }
-      width = MAX( width, node->rule_set->x + node->rule_set->width );
+      path_width = MAX( path_width, x + node->rule_set->width );
     }
   }
 
-  *width_p = width + border_width;
-  *height_p = height + border_width;
+  *width_p = path_width + border_width;
+  *height_p = path_height + border_width;
 }
 
 /*---------------------------------------------------------------------------*/
@@ -254,55 +291,80 @@ expose_path( canvas_t *canvas, rectangle_t *area )
 
   FOREACH( node, path_nodes )
   {
-    if (node->link_fs != NULL)
+    if (node->link_feat != NULL)
     {
       plus->x = border_width;
       plus->y = node->link_surf->y;
       draw_pos_string( plus, canvas );
       draw_pos_string( node->link_surf, canvas );
-      draw_pos_value( node->link_fs, canvas );
+      draw_pos_value( node->link_feat, canvas );
+      if (inline_path)
+      {
+	comma->x = node->link_surf->x + node->link_surf->width;
+	comma->y = node->link_surf->y;
+	draw_pos_string( comma, canvas );
+      }
     }
-    if (node->result_fs != NULL)
+    if (node->result_feat != NULL)
     {
-      x1 = node->rule_name->x;
-      x2 = x1 + node->rule_name->width + space_width;
-      y = node->rule_name->y + font_height + 1;
-      draw_pos_string( node->rule_name, canvas );
-      draw_lines( 2, x1, y, x2, y );
-      draw_lines( 3, 
-		  x2 - space_width, y - space_width / 2,
-		  x2, y, 
-		  x2 - space_width, y + space_width / 2 );
-      draw_pos_string( node->index, canvas );
+      if (node != (path_node_t *) path_nodes.first)
+      {
+	x1 = node->rule_name->x;
+	x2 = x1 + node->rule_name->width + space_width;
+	y = node->rule_name->y + font_height + 1;
+	draw_pos_string( node->rule_name, canvas );
+	draw_lines( 2, x1, y, x2, y );
+	draw_lines( 3, 
+		    x2 - space_width, y - space_width / 2,
+		    x2, y, 
+		    x2 - space_width, y + space_width / 2 );
+      }
+      if (path_show_state_indexes)
+	draw_pos_string( node->state_index, canvas );
       draw_pos_string( node->result_surf, canvas );
+      draw_pos_value( node->result_feat, canvas );
       draw_pos_string( node->rule_set, canvas );
-      draw_pos_value( node->result_fs, canvas );
+      if (inline_path)
+      {
+	comma->x = node->result_surf->x + node->result_surf->width;
+	comma->y = node->result_surf->y;
+	draw_pos_string( comma, canvas );
+	comma->x = node->result_feat->x + node->result_feat->width;
+	draw_pos_string( comma, canvas );
+      }
     }
   }
 }
 
 /*---------------------------------------------------------------------------*/
 
+static void
+unmark_tree_nodes( tree_node_t *node )
+/* Mark NODE and all its children and siblings *not* to be part of the path. */
+{
+  for (; node != NULL; node = node->sibling)
+  {
+    node->in_path = FALSE;
+    unmark_tree_nodes( node->child );
+  }
+}
+
+/*---------------------------------------------------------------------------*/
+
 static void
 free_path( void )
 {
   path_node_t *path_node;
-  int_t i;
 
   /* Clear old path nodes. */
   FOREACH_FREE( path_node, path_nodes )
   {
-    free_pos_string( &path_node->index );
-    free_pos_value( &path_node->link_fs );
-    free_pos_value( &path_node->result_fs );
+    free_pos_string( &path_node->state_index );
+    free_pos_value( &path_node->link_feat );
+    free_pos_value( &path_node->result_feat );
   }
 
-  /* Unmark all tree nodes. */
-  for (i = 0; i < node_count; i++)
-  {
-    if (nodes[i] != NULL) 
-      nodes[i]->in_path = FALSE;
-  }
+  unmark_tree_nodes( tree_nodes );
 }
 
 /*---------------------------------------------------------------------------*/
@@ -318,9 +380,18 @@ close_path( canvas_t *canvas )
 /*---------------------------------------------------------------------------*/
 
 static void 
-set_in_line( canvas_t *canvas, guint action, GtkWidget *item )
+set_inline_path( canvas_t *canvas, guint action, GtkWidget *item )
 {
-  in_line = GTK_CHECK_MENU_ITEM( item )->active;
+  inline_path = GTK_CHECK_MENU_ITEM( item )->active;
+  configure_canvas( canvas );
+}
+
+/*---------------------------------------------------------------------------*/
+
+static void 
+path_index_option( canvas_t *canvas, guint action, GtkWidget *item )
+{
+  path_show_state_indexes = GTK_CHECK_MENU_ITEM( item )->active;
   configure_canvas( canvas );
 }
 
@@ -328,12 +399,18 @@ set_in_line( canvas_t *canvas, guint action, GtkWidget *item )
 
 static GtkItemFactoryEntry path_items[] = 
 {
-  { "/Style/Inline", NULL, set_in_line, 0, "<ToggleItem>" },
+  { "/Style/Inline", NULL, set_inline_path, 0, "<ToggleItem>" },
+  { "/Path", NULL, NULL, 0, "<Branch>" },
+  { "/Path/Show State Indexes", NULL, path_index_option, 0, "<ToggleItem>" },
   { "/End States", NULL, NULL, 0, "<Branch>" },
-  { "/End States/Show First", "<Control>F", goto_state, FIRST_STATE, NULL },
-  { "/End States/Show Previous", "<Control>P", goto_state, PREV_STATE, NULL },
-  { "/End States/Show Next", "<Control>N", goto_state, NEXT_STATE, NULL },
-  { "/End States/Show Last", "<Control>L", goto_state, LAST_STATE, NULL }
+  { "/End States/Show First End State", "<Control>F", goto_state, FIRST_STATE, 
+    NULL },
+  { "/End States/Show Previous End State", "<Control>P", goto_state, PREV_STATE,
+    NULL },
+  { "/End States/Show Next End State", "<Control>N", goto_state, NEXT_STATE, 
+    NULL },
+  { "/End States/Show Last End State", "<Control>L", goto_state, LAST_STATE, 
+    NULL }
 };
 
 /*---------------------------------------------------------------------------*/
@@ -343,37 +420,40 @@ display_path( void )
 {
   tree_node_t *tree_node;
   path_node_t *path_node;
-  string_t index, string;
+  string_t state_index, string;
 
   free_path();
   
-  /* Append nodes to the path. */
+  /* Create the path nodes. We go backwards from the path end to the path 
+   * beginning, since this direction is easier to traverse. */
   for (tree_node = path_end; ; tree_node = tree_node->parent)
   {
     tree_node->in_path = TRUE;
     path_node = new_node( &path_nodes, sizeof( path_node_t ), LIST_START );
-    if (path_begin != NULL && tree_node->result_fs != NULL)
+    if (path_begin != NULL && tree_node->result_feat != NULL)
     {
-      index = int_to_string( tree_node->index );
-      string = concat_strings( "(", index, ")", NULL );
-      path_node->index = new_pos_string( string );
+      /* Create a new state in the path. */
+      state_index = int_to_string( tree_node->state_index );
+      string = concat_strings( "State ", state_index, ":", NULL );
+      path_node->state_index = new_pos_string( string );
       free_mem( &string );
-      free_mem( &index );
+      free_mem( &state_index );
       path_node->rule_name = new_pos_string( tree_node->rule_name );
       path_node->result_surf = new_pos_string( tree_node->result_surf );
-      set_scanner_input( tree_node->result_fs );
-      path_node->result_fs = parse_pos_value();
+      set_scanner_input( tree_node->result_feat );
+      path_node->result_feat = parse_pos_value();
       parse_token( EOF );
       set_scanner_input( NULL );
       path_node->rule_set = new_pos_string( tree_node->rule_set );
     }
     if (tree_node == path_begin) 
       break;
-    if (tree_node->link_fs != NULL)
+    if (tree_node->link_feat != NULL)
     {
+      /* Create a new link in the path. */
       path_node->link_surf = new_pos_string( tree_node->link_surf );
-      set_scanner_input( tree_node->link_fs );
-      path_node->link_fs = parse_pos_value();
+      set_scanner_input( tree_node->link_feat );
+      path_node->link_feat = parse_pos_value();
       parse_token( EOF );
       set_scanner_input( NULL );
     }
@@ -385,8 +465,12 @@ display_path( void )
   {
     path_canvas = create_canvas(
       "Malaga Path", "path.eps", &path_geometry, configure_path, expose_path,
-      close_path, NULL, TRUE, 
-      path_items, sizeof( path_items ) / sizeof( path_items[0] ) );
+      close_path, NULL, TRUE, path_items, ARRAY_LENGTH( path_items ) );
+    path_show_state_indexes = show_state_indexes;
+    if (path_show_state_indexes)
+      activate_menu_item( path_canvas, "/Path/Show State Indexes" );
+    if (inline_path)
+      activate_menu_item( path_canvas, "/Style/Inline" );
   }
   else
   {
@@ -399,17 +483,32 @@ display_path( void )
 /*---------------------------------------------------------------------------*/
 
 static void
-configure_node( canvas_t *canvas, tree_node_t *node, int_t x, int_t y,
-		int_t font_height )
+configure_node( canvas_t *canvas, tree_node_t *node, int_t x, int_t y )
 {
-  int_t sub_x;
+  int_t w, node_width;
   tree_node_t *subnode;
   bool_t is_first;
+  int_t font_height = get_font_height( canvas ); 
+  int_t space_width = get_space_width( canvas );
+
+  if (tree_show_state_indexes && node->state_index != -1)
+  {
+    config_pos_string( node->pos_state_index, canvas );
+    node_width = MAX( node->pos_state_index->width + space_width, 
+		      font_height );
+    node->pos_state_index->x 
+      = x + (node_width - node->pos_state_index->width) / 2;
+    node->pos_state_index->y = tree_height - font_height * 3 / 2 + 1;
+  }
+  else
+    node_width = font_height;
+
+  node->x = x + node_width / 2; 
+  node->y = y;
+  x += node_width;
+  tree_width = MAX( tree_width, x );
 
-  node->x = x; node->y = y;
-  
   is_first = TRUE;
-  width = MAX( width, x );
   for (subnode = node->child; subnode != NULL; subnode = subnode->sibling)
   {
     if (tree_mode == NO_DEAD_ENDS && subnode->type == BREAK_NODE) 
@@ -420,19 +519,18 @@ configure_node( canvas_t *canvas, tree_node_t *node, int_t x, int_t y,
     if (is_first) 
       is_first = FALSE;
     else 
-      height += (5 * font_height) / 2;
+      tree_height += (5 * font_height) / 2;
 
     config_pos_string( subnode->pos_link_surf, canvas );
     config_pos_string( subnode->pos_rule_name, canvas );
-    sub_x = x + font_height + MAX( subnode->pos_link_surf->width, 
-				   subnode->pos_rule_name->width );
-    subnode->pos_link_surf->x 
-      = (sub_x + x - subnode->pos_link_surf->width) / 2;
-    subnode->pos_link_surf->y = height - font_height - 1;
-    subnode->pos_rule_name->x 
-      = (sub_x + x - subnode->pos_rule_name->width) / 2;
-    subnode->pos_rule_name->y = height + 1;
-    configure_node( canvas, subnode, sub_x, height, font_height );
+
+    w = MAX( subnode->pos_link_surf->width, subnode->pos_rule_name->width );
+    w = MAX( w, space_width );
+    subnode->pos_link_surf->x = x + (w - subnode->pos_link_surf->width) / 2;
+    subnode->pos_link_surf->y = tree_height - font_height - 1;
+    subnode->pos_rule_name->x = x + (w - subnode->pos_rule_name->width) / 2;
+    subnode->pos_rule_name->y = tree_height + 1;
+    configure_node( canvas, subnode, x + w, tree_height );
   }
 }
 
@@ -444,26 +542,24 @@ configure_tree( canvas_t *canvas, int_t *width_p, int_t *height_p )
   int_t font_height = get_font_height( canvas );
   int_t border_width = get_border_width( canvas );
 
-  height = border_width;
-  width = border_width;
+  tree_width = tree_height = border_width;
 
   /* Configure result surface. */
   config_pos_string( result_surf, canvas );
   result_surf->x = border_width;
   result_surf->y = border_width;
-  width = MAX( width, border_width + result_surf->width );
-  height += font_height;
+  tree_width = MAX( tree_width, border_width + result_surf->width );
+  tree_height += font_height;
   if (tree_mode != RESULT_PATHS || tree_nodes->result_path)
   {
     /* Configure tree. */
-    height += 2 * font_height;
-    configure_node( canvas, tree_nodes, border_width + font_height / 2, height,
-		    font_height );
+    tree_height += 2 * font_height;
+    configure_node( canvas, tree_nodes, border_width, tree_height );
   }
 
   /* Return width and height. */
-  *width_p = width + font_height / 2 + border_width;
-  *height_p = height + font_height + border_width;
+  *width_p = tree_width + border_width;
+  *height_p = tree_height + font_height + border_width;
 }
 
 /*---------------------------------------------------------------------------*/
@@ -495,7 +591,7 @@ expose_node( canvas_t *canvas, tree_node_t *node, int_t font_height )
     else 
       set_color( BLACK );
     draw_lines( 3, x, old_y, x, subnode->y, subnode->x, subnode->y );
-    old_y = subnode->y;
+    old_y = subnode->y + 1;
     if (path_begin == NULL && subnode->in_path) 
       set_color( RED );
     draw_pos_string( subnode->pos_link_surf, canvas );
@@ -504,22 +600,18 @@ expose_node( canvas_t *canvas, tree_node_t *node, int_t font_height )
     expose_node( canvas, subnode, font_height );
   }
 
-  if (node->type != BREAK_NODE)
+  /* Draw the node itself. */
+  if (node->type == BREAK_NODE)
+  {
+    set_color( (node->in_path && path_begin != NULL ? RED : BLACK) );
+    draw_rectangle( x - radius1 , y - radius1, 2 * radius1, 2 * radius1 );
+  }
+  else
   {
     set_color( WHITE );
     draw_circle( TRUE, x, y, radius1 );
-  }
 
-  if (node->in_path && path_begin != NULL) 
-    set_color( RED );
-  else 
-    set_color( BLACK );
-
-  /* Finally, draw the node itself. */
-  if (node->type == BREAK_NODE)
-    draw_rectangle( x - radius1 , y - radius1, 2 * radius1, 2 * radius1 );
-  else
-  {
+    set_color( (node->in_path && path_begin != NULL ? RED : BLACK) );
     draw_circle( FALSE, x, y, radius1 );
     if (node->type == FINAL_NODE || node->type == UNFINAL_NODE)
       draw_circle( FALSE, x, y, radius2 );
@@ -529,12 +621,20 @@ expose_node( canvas_t *canvas, tree_node_t *node, int_t font_height )
       draw_lines( 2, x + radius1, y - radius1, x - radius1, y + radius1 );
     }
   }
+  
+  /* Draw the state index. */
+  if (tree_show_state_indexes && node->state_index != -1)
+  {
+    set_color( BLUE );
+    draw_pos_string( node->pos_state_index, canvas );
+  }
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void
 expose_tree( canvas_t *canvas, rectangle_t *area )
+/* Expose AREA on tree CANVAS. */
 {
   int_t font_height = get_font_height( canvas );
 
@@ -553,17 +653,24 @@ expose_tree( canvas_t *canvas, rectangle_t *area )
 
 static bool_t
 in_circle( int_t x, int_t y, int_t circle_x, int_t circle_y, int_t radius )
+/* Return TRUE if position (X,Y) lies in disc with center (CIRCLE_X, CIRCLE_Y)
+ * and RADIUS. */ 
 {
   int_t diff_x = x - circle_x;
   int_t diff_y = y - circle_y;
   
-  return (diff_x * diff_x + diff_y * diff_y <= radius * radius);
+  /* Taking DIFF_X or DIFF_Y to the 2nd power may result in an overflow,
+   * so we will check first whether DIFF_X and DIFF_Y are small enough. */
+  return (ABS( diff_x ) < radius && ABS( diff_y ) < radius
+	  && diff_x * diff_x + diff_y * diff_y <= radius * radius);
 }
 
 /*---------------------------------------------------------------------------*/
 
 static bool_t
 in_pos_string( int_t x, int_t y, pos_string_t *pos_string, int_t height )
+/* Return TRUE if position (X,Y) lies in bounding box of POS_STRING, using font
+ * height HEIGHT. */
 {
   return (x >= pos_string->x 
 	  && x < pos_string->x + pos_string->width
@@ -573,6 +680,20 @@ in_pos_string( int_t x, int_t y, pos_string_t *pos_string, int_t height )
 
 /*---------------------------------------------------------------------------*/
 
+static bool_t
+in_rectangle( int_t x, int_t y, 
+	      int_t rect_x, int_t rect_y, int_t width, int_t height )
+/* Return TRUE if position (X,Y) lies in rectangle with upper left corner 
+ * (RECT_X, RECT_Y), width WIDTH and height HEIGHT. */
+{
+  return (x >= rect_x 
+	  && x <= rect_x + width 
+	  && y >= rect_y
+	  && y <= rect_y + height);
+}
+
+/*---------------------------------------------------------------------------*/
+
 static void 
 set_tree( canvas_t *canvas, guint mode )
 /* Set the tree's display mode to MODE. */
@@ -591,6 +712,7 @@ set_tree( canvas_t *canvas, guint mode )
 
 static void 
 goto_state( canvas_t *canvas, guint action )
+/* Display a certain state, according to ACTION, in the path window. */
 {
   int_t i = 0;
   bool_t inverse = FALSE; /* TRUE if we search in inverse direction. */
@@ -604,22 +726,22 @@ goto_state( canvas_t *canvas, guint action )
       i = 0; 
       break;
     case PREV_STATE:
-      i = path_begin->index - 1;
+      i = path_begin->state_index - 1;
       inverse = TRUE;
       break;
     case NEXT_STATE:
-      i = path_begin->index + 1;
+      i = path_begin->state_index + 1;
       break;
     case LAST_STATE: 
-      i = node_count - 1; 
+      i = state_count - 1; 
       inverse = TRUE; 
       break;
     } 
-    while (i >= 0 && i < node_count)
+    while (i >= 0 && i < state_count)
     {
-      if (nodes[i] != NULL && nodes[i]->type == FINAL_NODE)
+      if (states[i] != NULL && states[i]->type == FINAL_NODE)
       {
-	path_begin = path_end = nodes[i];
+	path_begin = path_end = states[i];
 	make_visible( tree_canvas, path_begin->x, path_begin->y );
 	display_path();
 	break;
@@ -634,76 +756,133 @@ goto_state( canvas_t *canvas, guint action )
 
 /*---------------------------------------------------------------------------*/
 
-static bool_t
-mouse_event(  canvas_t *canvas, int_t x, int_t y, int_t button )
-/* Called if mouse has moved to position (X,Y). */
+static tree_node_t *
+tree_node_at_position( tree_node_t *node, int_t x, int_t y, int_t font_height, 
+		       element_t *element_p)
 {
-  tree_node_t *node, *node2, *new_begin, *new_end;
-  int_t i;
-  int_t font_height = get_font_height( canvas );
+  tree_node_t *subnode;
 
-  for (i = 0; i < node_count; i++)
+  if (node == NULL) 
+    return NULL;
+  for (; node != NULL; node = node->sibling)
   {
-    node = nodes[i];
-    if (node == NULL) 
-      continue;
     if (tree_mode == NO_DEAD_ENDS && node->type == BREAK_NODE) 
       continue;
     if (tree_mode == RESULT_PATHS && ! node->result_path) 
       continue;
-    if (in_circle( x, y, node->x, node->y, font_height / 2 )
-	&& node->type != BREAK_NODE)
+    if ((node->type == BREAK_NODE 
+	 && in_rectangle( x, y, 
+			  node->x - font_height / 2, node->y - font_height / 2,
+			  font_height, font_height ))
+	|| (node->type != BREAK_NODE 
+	    && in_circle( x, y, node->x, node->y, font_height / 2 ) ))
     {
-      if (button == 0) 
-	set_cursor( canvas, TRUE );
-      else
-      {
-	if (button == 1 || button == 2) 
-	  new_begin = node;
-	else 
-	  new_begin = path_begin;
-	if (button == 2 || button == 3) 
-	  new_end = node;
-	else 
-	  new_end = path_end;
-	node2 = new_end; 
-	while (node2 != NULL && node2 != new_begin) 
-	  node2 = node2->parent;
-	if (node2 == NULL) 
-	  new_begin = new_end = node;
-	path_begin = new_begin;
-	path_end = new_end;
-	display_path();
-      }
-      return TRUE;
+      *element_p = ELEMENT_NODE;
+      return node;
     }
     if (in_pos_string( x, y, node->pos_link_surf, font_height ))
     {
-      if (button == 0) 
-	set_cursor( canvas, TRUE );
-      else
-      {
-	path_begin = NULL;
-	path_end = node;
-	display_path();
-      }
-      return TRUE;
+      *element_p = ELEMENT_LINK;
+      return node;
     }
     if (in_pos_string( x, y, node->pos_rule_name, font_height ))
-    {
-      if (button == 0) 
-	set_cursor( canvas, TRUE );
-      else
-      {
-	path_begin = node->parent;
-	path_end = node;
-	display_path();
-      }
-      return TRUE;
+    { 
+      *element_p = ELEMENT_RULE;
+      return node;
     }
+    subnode = tree_node_at_position( node->child, x, y, font_height, 
+				     element_p );
+    if (subnode != NULL)
+      return subnode;
   }
-  set_cursor( canvas, FALSE );
-  return FALSE;
+  return NULL;
+}
+
+/*---------------------------------------------------------------------------*/
+
+static void
+display_node( canvas_t *canvas, guint action )
+{
+  tree_node_t *node;
+
+  if (action == PATH_BEGIN)
+    path_begin = popup_menu_node;
+  else if (action == FROM_ROOT)
+    path_begin = tree_nodes;
+  if (action == PATH_END || action == FROM_ROOT)
+    path_end = popup_menu_node;
+
+  /* Check whether there is a part from PATH_BEGIN to PATH_END. */ 
+  node = path_end; 
+  while (node != NULL && node != path_begin) 
+    node = node->parent;
+  if (node == NULL) 
+    path_begin = path_end = popup_menu_node;
+  
+  display_path();
+}
+
+/*---------------------------------------------------------------------------*/
+
+static bool_t
+mouse_event(  canvas_t *canvas, int_t x, int_t y, int_t button )
+/* Called if mouse has moved to position (X,Y). */
+{
+  tree_node_t *node;
+  element_t element;
+
+  node = tree_node_at_position( tree_nodes, x, y, get_font_height( canvas ),
+				&element );
+  if (node == NULL)
+  {
+    set_cursor( canvas, FALSE );
+    return FALSE;
+  }
+
+  /* No button pressed: we should just update the cursor. */
+  if (button == 0)
+  {
+    set_cursor( canvas, TRUE );
+    return TRUE;
+  }
+
+  switch (element)
+  {
+  case ELEMENT_NODE:
+    if (button == 1) 
+    {
+      path_begin = path_end = node;
+      display_path();
+    }
+    else if (button == 3)
+    {
+      popup_menu_node = node;
+      popup_menu( canvas );
+    }
+    break;
+
+  case ELEMENT_LINK:
+    path_begin = NULL;
+    path_end = node;
+    display_path();
+    break;
+
+  case ELEMENT_RULE:
+    path_begin = node->parent;
+    path_end = node;
+    display_path();
+    break;
+  }
+  return TRUE;
+}
+
+/*---------------------------------------------------------------------------*/
+
+static void 
+tree_index_option( canvas_t *canvas, guint action, GtkWidget *item )
+{
+  tree_show_state_indexes = GTK_CHECK_MENU_ITEM( item )->active;
+  configure_canvas( canvas );
 }
 
 /*---------------------------------------------------------------------------*/
@@ -714,38 +893,67 @@ static GtkItemFactoryEntry tree_items[] =
   { "/Tree/Full Tree", NULL, set_tree, FULL_TREE, "<RadioItem>" },
   { "/Tree/No Dead Ends", NULL, set_tree, NO_DEAD_ENDS, "/Tree/Full Tree" },
   { "/Tree/Complete Paths", NULL, set_tree, RESULT_PATHS, "/Tree/Full Tree" },
+  { "/Tree/sep1", NULL, NULL, 0, "<Separator>" },
+  { "/Tree/Show State Indexes", NULL, tree_index_option, 0, "<ToggleItem>" },
   { "/End States", NULL, NULL, 0, "<Branch>" },
-  { "/End States/Show First", "<Control>F", goto_state, FIRST_STATE, NULL },
-  { "/End States/Show Previous", "<Control>P", goto_state, PREV_STATE, NULL },
-  { "/End States/Show Next", "<Control>N", goto_state, NEXT_STATE, NULL },
-  { "/End States/Show Last", "<Control>L", goto_state, LAST_STATE, NULL }
+  { "/End States/Show First End State", "<Control>F", goto_state, FIRST_STATE, 
+    NULL },
+  { "/End States/Show Previous End State", "<Control>P", goto_state, PREV_STATE,
+    NULL },
+  { "/End States/Show Next End State", "<Control>N", goto_state, NEXT_STATE, 
+    NULL },
+  { "/End States/Show Last End State", "<Control>L", goto_state, LAST_STATE, 
+    NULL }
 };
 
 /*---------------------------------------------------------------------------*/
 
+static GtkItemFactoryEntry popup_items[] = 
+{
+  { "/Set Path Begin", NULL, display_node, PATH_BEGIN, "<Item>" },
+  { "/Set Path End", NULL, display_node, PATH_END, "<Item>" },
+  { "/Show Path From Root", NULL, display_node, FROM_ROOT, "<Item>" }
+};
+
+/*---------------------------------------------------------------------------*/
+
+static void
+free_tree_nodes( tree_node_t *node )
+/* Free *NODE_P and all its siblings and children. Set *NODE_P to NULL. */
+{
+  tree_node_t *next;
+
+  for (; node != NULL; node = next)
+  {
+    /* Remove node from state index. */
+    if (node->state_index != -1)
+      states[ node->state_index ] = NULL;
+
+    free_tree_nodes( node->child );
+    free_mem( &node->rule_name );
+    free_mem( &node->link_surf );
+    free_mem( &node->link_feat );
+    free_mem( &node->result_surf );
+    free_mem( &node->result_feat );
+    free_mem( &node->rule_set );
+    free_pos_string( &node->pos_link_surf );
+    free_pos_string( &node->pos_rule_name );
+    free_pos_string( &node->pos_state_index );
+    next = node->sibling;
+    free( node );
+  }
+}
+
+/*---------------------------------------------------------------------------*/
+
 static void
 free_tree( void )
 {
-  int_t i;
-
   /* Clear old variables. */
   free_pos_string( &result_surf );
   free_pos_string( &plus );
-  for (i = 0; i < node_count; i++) 
-  {
-    if (nodes[i] != NULL)
-    {
-      free_mem( &nodes[i]->rule_name );
-      free_mem( &nodes[i]->link_surf );
-      free_mem( &nodes[i]->link_fs );
-      free_mem( &nodes[i]->result_surf );
-      free_mem( &nodes[i]->result_fs );
-      free_mem( &nodes[i]->rule_set );
-      free_pos_string( &nodes[i]->pos_link_surf );
-      free_pos_string( &nodes[i]->pos_rule_name );
-      free_mem( &nodes[i] );
-    } 
-  }
+  free_pos_string( &comma );
+  free_tree_nodes( tree_nodes );
   tree_nodes = NULL;
 }
 
@@ -769,8 +977,9 @@ read_tree( void )
   string_t line; /* A line of input from STDIN. */
   string_t line_p; /* Part of LINE which is yet to parse. */
   string_t type; /* Node type. */
+  string_t state_index;
   tree_node_t *node;
-  int_t parent_index, i;
+  int_t parent_state_index, i;
   tree_node_t **node_p;
 
   if (path_canvas != NULL) 
@@ -778,10 +987,10 @@ read_tree( void )
   free_tree();
 
   /* Allocate new nodes array. */
-  if (node_count == 0)
+  if (state_count == 0)
   {
-    node_count = 100;
-    nodes = new_vector( sizeof( tree_node_t * ), node_count );
+    state_count = 100;
+    states = new_vector( sizeof( tree_node_t * ), state_count );
   }
 
   /* Read new input surface. */
@@ -800,7 +1009,7 @@ read_tree( void )
 
     node = new_mem( sizeof( tree_node_t ) );
     line_p = line;
-    node->index = parse_int( &line_p );
+    node->state_index = parse_int( &line_p );
     type = parse_word( &line_p );
     if (strcmp_no_case( type, "inter" ) == 0) 
       node->type = INTER_NODE;
@@ -815,27 +1024,36 @@ read_tree( void )
     else 
       complain( "Unknown node type \"%s\".", type );
     free_mem( &type );
-    parent_index = parse_int( &line_p );
+    parent_state_index = parse_int( &line_p );
     node->rule_name = parse_word( &line_p );
     node->link_surf = parse_optional_value( &line_p );
-    node->link_fs = parse_optional_value( &line_p );
+    node->link_feat = parse_optional_value( &line_p );
     node->result_surf = parse_optional_value( &line_p );
-    node->result_fs = parse_optional_value( &line_p );
+    node->result_feat = parse_optional_value( &line_p );
     node->rule_set = parse_word( &line_p );
     parse_end( &line_p );
     free_mem( &line );
 
     node->pos_link_surf = new_pos_string( node->link_surf );
     node->pos_rule_name = new_pos_string( node->rule_name );
+    if (node->state_index != -1)
+    {
+      state_index = int_to_string( node->state_index );
+      node->pos_state_index = new_pos_string( state_index );
+      free_mem( &state_index );
+    }
 
     /* Find parent node. */
-    if (parent_index == -1) 
-      tree_nodes = node;
-    else if (parent_index >= node_count || nodes[ parent_index ] == NULL)
+    if (parent_state_index == -1) 
+      tree_nodes = node; /* This is the root node. */
+    else if (parent_state_index < 0 || parent_state_index >= state_count 
+	     || states[ parent_state_index ] == NULL)
+    {
       complain( "Display data corrupted." );
+    }
     else 
     { 
-      node->parent = nodes[ parent_index ];
+      node->parent = states[ parent_state_index ];
 
       /* Add NODE to be a child of NODE->PARENT. */
       node_p = &node->parent->child;
@@ -844,15 +1062,16 @@ read_tree( void )
       *node_p = node;
     }
 
-    /* Add node to the indexes. */
-    if (node->index >= node_count)
+    /* Add node to the STATES array. */
+    if (node->state_index >= state_count)
     {
-      renew_vector( &nodes, sizeof( tree_node_t * ), 2 * node->index );
-      for (i = node_count; i < 2 * node->index; i++) 
-	nodes[i] = NULL;
-      node_count = 2 * node->index;
+      renew_vector( &states, sizeof( tree_node_t * ), 2 * node->state_index );
+      for (i = state_count; i < 2 * node->state_index; i++) 
+	states[i] = NULL;
+      state_count = 2 * node->state_index;
     }
-    nodes[ node->index ] = node;
+    if (node->state_index != -1)
+      states[ node->state_index ] = node;
 
     /* If this is a final node, mark it and all its predecessors. */
     if (node->type == FINAL_NODE) 
@@ -866,13 +1085,17 @@ read_tree( void )
     complain( "Missing root node in tree." );
 
   plus = new_pos_string( "+" );
+  comma = new_pos_string( "," );
 
   if (tree_canvas == NULL) 
   {
     tree_canvas = create_canvas( 
-      "Malaga Tree", "tree.eps", &tree_geometry, configure_tree, expose_tree,
-       close_tree, mouse_event, 
-       FALSE, tree_items, sizeof( tree_items ) / sizeof( tree_items[0] ) );
+      "Malaga Tree", "tree.eps", &tree_geometry, configure_tree, expose_tree, 
+      close_tree, mouse_event, FALSE, tree_items, ARRAY_LENGTH( tree_items ) );
+    set_popup_menu( tree_canvas, popup_items, ARRAY_LENGTH( popup_items ) );
+    tree_show_state_indexes = show_state_indexes;
+    if (tree_show_state_indexes )
+      activate_menu_item( tree_canvas, "/Tree/Show State Indexes" );
   }
   else
   {
